unit User;

{This module is a good place to put user additions to Image. You will need }
{to uncomment the call to InitUser in Image.p.}


interface

	uses
		QuickDraw, Palettes, PrintTraps, globals, Utilities, Graphics, Functions, Analysis;


	procedure InitUser;
	procedure DoUserCommand1;
	procedure DoUserCommand2;
	procedure DoUserMenuEvent (MenuItem: integer);
	procedure TestFindPeak (event: EventRecord);
	procedure UserMacroCode (CodeNumber: integer; Param1, Param2, Param3: extended);


implementation

{User global variables go here.}
	var
		color: integer;
		SaveInfo: InfoPtr;
		radius, slope: extended;


	procedure InitUser;
	begin
		UserMenuH := GetMenu(UserMenu);
		InsertMenu(UserMenuH, 0);
		DrawMenuBar;
{Additional user initialization code goes here.}
	end;


	procedure DrawDot (row, column, RowOffset, ColumnOffset: integer; big: boolean);
		var
			h, v: integer;
	begin
		if big then begin
				for h := -1 to 1 do
					for v := -1 to 1 do
						PutPixel(column * 16 + ColumnOffset * 4 + h + 16, row * 16 + RowOffset * 4 + v + 16, color)
			end
		else
			PutPixel(column * 16 + ColumnOffset * 4 + 16, row * 16 + RowOffset * 4 + 16, color);
	end;

	procedure DrawNeighborhood (i, row, column: integer);

	begin
		DrawDot(row, column, 0, 0, BitAnd(i, 1) = 1);
		DrawDot(row, column, 0, 1, BitAnd(i, 2) = 2);
		DrawDot(row, column, 0, 2, BitAnd(i, 4) = 4);
		DrawDot(row, column, 1, 2, BitAnd(i, 8) = 8);
		DrawDot(row, column, 2, 2, BitAnd(i, 16) = 16);
		DrawDot(row, column, 2, 1, BitAnd(i, 32) = 32);
		DrawDot(row, column, 2, 0, BitAnd(i, 64) = 64);
		DrawDot(row, column, 1, 0, BitAnd(i, 128) = 128);
		DrawDot(row, column, 1, 1, true);
	end;


	procedure SetColor (i: integer);
{Color neighborhoods to show which ones would be removed on the first pass(150), second pass(100),}
{or either pass(200) when using the Zhang and Suen thinning algorithm(CACM, Mar. 1984,236-239).}
		var
			p2, p3, p4, p5, p6, p7, p8, p9, A, B: integer;
	begin
		p2 := bsr(band(i, 2), 1);
		p3 := bsr(band(i, 4), 2);
		p4 := bsr(band(i, 8), 3);
		p5 := bsr(band(i, 16), 4);
		p6 := bsr(band(i, 32), 5);
		p7 := bsr(band(i, 64), 6);
		p8 := bsr(band(i, 128), 7);
		p9 := band(i, 1);
		A := 0;
		if (p2 = 0) and (p3 = 1) then
			A := A + 1;
		if (p3 = 0) and (p4 = 1) then
			A := A + 1;
		if (p4 = 0) and (p5 = 1) then
			A := A + 1;
		if (p5 = 0) and (p6 = 1) then
			A := A + 1;
		if (p6 = 0) and (p7 = 1) then
			A := A + 1;
		if (p7 = 0) and (p8 = 1) then
			A := A + 1;
		if (p8 = 0) and (p9 = 1) then
			A := A + 1;
		if (p9 = 0) and (p2 = 1) then
			A := A + 1;
		B := p2 + p3 + p4 + p5 + p6 + p7 + p8 + p9;
		color := 255;
		if A = 1 then
			if (B >= 2) and (B <= 6) then begin
					if ((p2 * p4 * p6 = 0) and (p4 * p6 * p8 = 0)) and ((p2 * p4 * p8 = 0) and (p2 * p6 * p8 = 0)) then
						color := 200
					else if (p2 * p4 * p6 = 0) and (p4 * p6 * p8 = 0) then
						color := 150
					else if (p2 * p4 * p8 = 0) and (p2 * p6 * p8 = 0) then
						color := 100;
				end;
	end;


	procedure DoUserCommand1;
{Generates a table showing all possible 3x3 neighborhoods. This table is used}
{ for making up the "fate table" used by the Skeletonize command and the Wand tool.}
		var
			row, column, index: integer;
	begin
		row := 0;
		column := 0;
		if NewPicWindow('Fate Table', 600, 200) then
			for index := 0 to 255 do begin
					SetColor(index);
					DrawNeighborhood(index, row, column);
					column := column + 1;
					if column = 32 then begin
							row := row + 1;
							column := 0;
						end;
				end;
	end;


	function FindPeak (x, y, MinValue: integer): boolean;
		const
			MinSpacing = 4;
		var
			delta, angle: extended;
			v, dx, dy, i, v2, maxv2, x2, y2, v2count, nSamples: integer;
			sample: LineType;
			minlower, count, nLower, maxCount: integer;
			ControlKey, isSpot: boolean;
			mask: rect;
	begin
		FindPeak := false;
		v := MyGetPixel(x, y);
		ControlKey := ControlKeyDown;
		if ControlKey then begin
				v2count := 1;
				maxv2 := round((1.0 - slope) * v);
				User1^[v2count] := maxv2;
				npoints := 0;
				nLengths := 0;
				nAngles := 0;
				mCount := 1;
				mCount2 := 1;
				Measurements := [User1M];
			end;
		if v < MinValue then
			exit(FindPeak);
		if v <= MyGetPixel(x + 1, y) then
			exit(FindPeak);
		if v <= MyGetPixel(x + 1, y + 1) then
			exit(FindPeak);
		if v <= MyGetPixel(x, y + 1) then
			exit(FindPeak);
		if v <= MyGetPixel(x - 1, y + 1) then
			exit(FindPeak);
		if v < MyGetPixel(x - 1, y) then
			exit(FindPeak);
		if (v < MyGetPixel(x - 1, y - 1)) then
			exit(FindPeak);
		if v < MyGetPixel(x, y - 1) then
			exit(FindPeak);
		if v < MyGetPixel(x + 1, y - 1) then
			exit(FindPeak);
		nSamples := round(4 * radius);
		delta := 2.0 * pi / nsamples;
		angle := 0.0;
		maxv2 := round((1.0 - slope) * v);
		for i := 1 to nSamples do begin
				dx := round(radius * cos(angle));
				dy := round(radius * sin(angle));
				sample[i] := MyGetPixel(x + dx, y + dy);
				if ControlKey then begin
						v2count := v2count + 1;
						User1^[v2count] := v2;
						mcount := v2count;
						Info := SaveInfo;
						x2 := x + dx;
						y2 := y + dy;
						PutPixel(x2, y2, 0);
						SetRect(mask, x2, y2, x2 + 1, y2 + 1);
						UpdateScreen(mask);
						Info := UndoInfo;
					end;
				angle := angle + delta;
			end;
		minLower := round(0.677 * nsamples);
		isSpot := false;
		count := 0;
		i := 1;
		nLower := 0;
		maxCount := nSamples + minLower;
		repeat
			if sample[i] <= maxv2 then
				nLower := nLower + 1
			else
				nLower := 0;
			isSpot := nLower >= minLower;
			i := i + 1;
			if i > nSamples then
				i := 1;
			count := count + 1;
		until isSpot or (count = maxCount);
		if isSpot then begin
				info := SaveInfo;
				with info^ do begin
						SetRect(RoiRect, x - MinSpacing + 1, y - MinSpacing + 1, x + MinSpacing, y + MinSpacing);
						with RoiRect do begin
								if left < 0 then
									left := 0;
								if top < 0 then
									top := 0;
								if right > PicRect.right then
									right := PicRect.right;
								if bottom > PicRect.bottom then
									bottom := PicRect.bottom;
							end;
						GetRectHistogram;
						isSpot := histogram[0] = 0;
					end; {with}
				Info := UndoInfo;
			end;
		FindPeak := isSpot;
	end;


	procedure TestFindPeak (event: EventRecord);
		var
			loc: point;
			x, y: integer;
	begin
		if NoUndo then
			exit(TestFindPeak);
		SetPort(GrafPtr(info^.osPort));
		loc := event.where;
		ScreenToOffscreen(loc);
		with loc do begin
				x := h;
				y := v;
			end;
		SetupUndoInfoRec;
		SaveInfo := info;
		Info := UndoInfo;
		if FindPeak(x, y, 100) then
			beep;
		Info := SaveInfo;
		UpdateList;
	end;


	procedure CountSpots (MinValue, RadiusP, SlopeP: extended);
		var
			x, y, i, iMinValue: integer;
			AutoSelectAll: boolean;
			srect, mask: rect;
			count: LongInt;
			t: FateTable;
	begin
		if NotRectangular or NotInBounds or NoUndo then
			exit(CountSpots);
		iMinValue := round(MinValue);
		if iMinValue < 10 then
			iMinValue := 10;
		if iMinValue > 150 then
			iMinValue := 150;
		radius := RadiusP;
		if radius = 0.0 then
			radius := 6.0;
		if radius < 1.0 then
			radius := 1.0;
		if radius > 50.0 then
			radius := 50.0;
		slope := SlopeP;
		if slope = 0.0 then
			slope := 0.2;
		if slope < 0.05 then
			slope := 0.05;
		if slope > 0.95 then
			slope := 0.95;
		AutoSelectAll := not Info^.RoiShowing;
		if AutoSelectAll then
			SelectAll(true);
		ShowWatch;
		SetupUndo;
		WhatToUndo := UndoEdit;
		SetupUndoInfoRec;
		SaveInfo := Info;
		srect := info^.roiRect;
		KillRoi;
		ChangeValues(0, 0, 1);
		info := UndoInfo;
		count := 0;
		with srect do
			for y := top to bottom - 1 do begin
					if CommandPeriod then begin
							beep;
							Info := SaveInfo;
							leave;
						end;
					for x := left to right - 1 do
						if FindPeak(x, y, iMinValue) then begin
								count := count + 1;
								Info := SaveInfo;
								PutPixel(x, y, 0);
								PutPixel(x - 1, y, 0);
								PutPixel(x - 1, y - 1, 0);
								PutPixel(x, y - 1, 0);
								SetRect(mask, x - 1, y - 1, x + 1, y + 1);
								UpdateScreen(mask);
								Info := UndoInfo;
								ShowMessage(concat(long2str(y), '  ', long2str(count)));
							end;
				end;
		Info := SaveInfo;
		if count < MaxMeasurements then begin
				UnsavedResults := false;
				ResetCounter;
				for i := 1 to count do
					ClearResults(i);
				mCount := count;
				UpdateList;
				ShowValues;
			end
		else
			PutMessage('"Max Measurements" is too small.');
		ShowMessage(concat('Count=', long2str(count), cr, 'Threshold=', long2str(iMinValue)));
	end;


	procedure ShowNoCodeMessage;
	begin
		PutMessage('Requires user written Think Pascal routine. ');
	end;


	procedure DoUserCommand2;
	begin
		ShowNoCodeMessage
	end;


	procedure DoUserMenuEvent (MenuItem: integer);
	begin
		case MenuItem of
			1: 
				DoUserCommand1;
			2: 
				DoUserCommand2;
		end;
	end;


	procedure UserMacroCode (CodeNumber: integer; Param1, Param2, Param3: extended);
	begin
		case CodeNumber of
			1: 
				ShowNoCodeMessage;
			2: 
				ShowNoCodeMessage;
			3: 
				ShowNoCodeMessage;
			4: 
				ShowNoCodeMessage;
			5: 
				CountSpots(param1, param2, param3);
		end;
	end;
end.